
/* fdisk.c -- Atari partition table manipulator for Linux (and GEMDOS).
 *
 * Copyright (C) 1995 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
 *
 * Some parts of this Atari version of fdisk have been taken from the sources
 * for the original PC Linux fdisk by A. V. Le Blanc (LeBlanc@mcc.ac.uk) to
 * keep the interface between the two as consistent as possible.
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation: either version 2 or
 * (at your option) any later version.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <setjmp.h>
#include <sys/ioctl.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/fs.h>

#define	VERSION		"0.1"

#define DEFAULT_DEVICE		"/dev/sda"
#define ALTERNATE_DEVICE	"/dev/hda"
#define ALTERNATE2_DEVICE	"/dev/ada"
#define LINE_LENGTH			80
#define MAXIMUM_PARTS		60
#define SECTOR_SIZE			512

enum failure {
	usage, unable_to_open, unable_to_read, unable_to_seek,
	unable_to_write, out_of_memory, general
};

char	*disk_device = DEFAULT_DEVICE,	/* sda, unless specified */
		*line_ptr,						/* interactive input */
		line_buffer[LINE_LENGTH],
		changed[MAXIMUM_PARTS],		/* marks changed buffers */
		buffer[SECTOR_SIZE],		/* first four partitions */
		*buffers[MAXIMUM_PARTS]		/* pointers to buffers */
		= {buffer, buffer, buffer, buffer};

int	fd,					/* the disk */
	ext_index,			/* the prime extended partition */
	listing = 0,		/* no aborts for fdisk -l */
	size_flag = 0,
	partitions = 4;		/* maximum partition + 1 */


/* partition format structures on disk */

#define PID_EQ(a,b)		(!memcmp( (a), (b), 3 ))

struct partition {
	unsigned char	flag;			/* bit 0: valid */
#define	PART_FLAG_VALID		0x01
	char 			id[3];			/* "GEM", "BGM", "XGM", or other */
	unsigned long	start;			/* start of partition */
	unsigned long	size;			/* length of partition */
};

struct rootsector {
	char 					unused[0x156];	/* room for boot code */
	struct partition_info	icdpart[8];		/* info for ICD-partitions 5..12 */
	char 					unused2[0xc];
	unsigned long			hd_size;		/* size of disk in blocks */
	struct partition_info	part[4];
	unsigned long			bsl_st;			/* start of bad sector list */
	unsigned long			bsl_cnt;		/* length of bad sector list */
	unsigned short			checksum;		/* checksum for bootable disks */
};

/* in-core representation of a partition */

typedef struct _PARTITION {
	unsigned long	start;			/* abs. start of partition */
	unsigned long	size;			/* size of partition */
	int				flag;			/* boot flags */
	char			id[4];			/* ID, 0-terminated */
	unsigned long	rootsec;		/* sector# where this entry comes from;
									 * if != 0: an extended part.  */
	struct _PARTITION *next;		/* pointer to first subpart. if ext. part
									 * or to next subpart. if subpart. */
} PARTITION;

PARTITION part_table[MAXIMUM_PARTS];

int				using_ICD_format;
unsigned long	hd_size;
int				rootsec_bootable;


struct partition_IDs {
	char id[4];
	char *name;
} partition_IDs[] = {
	{ "BGM", "GEMDOS (> 16 MB)" },
	{ "GEM", "GEMDOS (< 16 MB)" },
	{ "LNX", "Linux" },
	{ "MAC", "Mac HFS" },
	{ "MIX", "Minix" },
	{ "RAW", "unspecified" }
	{ "SWP", "Swap" },
	{ "UNX", "Atari System V Unix" },
};

struct bootflags {
	unsigned char	flag;
	char			*name;
} bootflags[] = {
	{ 0x80, "GEMDOS" },
	{ 0x40, "Atari System V" },
	{ 0x20, "(not assigned)" },
	{ 0x10, "(not assigned)" },
	{ 0x08, "(not assigned)" }
};

jmp_buf listingbuf;


void fatal( enum failure why )

{
	char	error[LINE_LENGTH], *message = error;

	if (listing) {
		dev_deinit();
		longjmp( listingbuf, 1 );
	}

	switch( why ) {
	  case usage:
		message = "Usage: fdisk [-l] [-v] [-s /dev/hdxn] [/dev/hdx]\n";
		break;
	  case unable_to_open:
		sprintf(error, "Unable to open %s\n", disk_device);
		break;
	  case unable_to_read:
		sprintf(error, "Unable to read %s\n", disk_device);
		break;
	  case unable_to_seek:
		sprintf(error, "Unable to seek on %s\n", disk_device);
		break;
	  case unable_to_write:
		sprintf(error, "Unable to write %s\n", disk_device);
		break;
	  case out_of_memory:
		message = "Unable to allocate any more memory\n";
		break;
	  default:
		message = "Fatal error\n";
	}

	fputc( '\n', stderr );
	fputs( message, stderr );
	exit( 1 );
}

void menu(void)
{
	puts( "Command action\n"
		  "   a   set bootable flags\n"
		  "   d   delete a partition\n"
		  "   g   grow a partition\n"
		  "   l   list known partition types\n"
		  "   m   print this menu\n"
		  "   n   add a new partition\n"
		  "   p   print the partition table\n"
		  "   q   quit without saving changes\n"
		  "   s   shrink a partition\n"
		  "   t   change a partition's ID\n"
		  "   v   verify the partition table\n"
		  "   w   write table to disk and exit\n"
		  "   x   move a partition\n"
	);
}



/* ---------------------------------------------------------------------- */
/*						 Reading the partition table					  */


void diskpart2PART( struct partition *dp, unsigned long sec, PARTITION *cp )

{
	cp->start   = sec + dp->start;
	cp->size    = dp->size;
	cp->flag    = dp->flag & ~PART_FLAG_VALID;
	cp->rootsec = sec;

	memcpy( cp->id, dp->id, 3 );
	cp->id[3] = 0;
}


void check_n_partitions( void )

{
	if (partitions == MAX_PARTITIONS) {
		fprintf( stderr, "Too much (more than %d) partitions on this disk!\n",
				 MAX_PARTITIONS );
		fatal( general );
	}
}


void read_extended( struct partition *pi )

{	char				buffer[512];
	struct rootsector	*xrs;
	unsigned long		partsect, extensect;

	partsect = extensect = pi->start;

	while( 1 ) {

		if (dev_read( buffer, partsect ))
			fatal( unable_to_read );
		xrs = (struct rootsector *)buffer;
		
		if (!(xrs->part[0].flag & PART_FLAG_VALID)) {
			fprintf( stderr, "Partition format error:\n"
					 "  First sub-partition in extended partition is "
					 "not valid!\n" );
			fatal( general );
		}

		check_n_partitions();
		diskpart2PART( &xrs->part[0], partsect, &part_table[partition] );
		++partition;
		
		if (!(xrs->part[1].flg & PART_FLAG_VALID))
			/* end of linked partition list */
			break;

		if (PID_EQ( xrs->part[1].id, "XGM" )) {
			fprintf( stderr, "Partition format error:\n"
					 "  ID of extended partion is not XGM!\n" );
			fatal( general );
		}
		
		partsect = xrs->part[1].start + extensect;
	}
}


/* read the partition table */

void get_boot(void)

{	char				buffer[512];
	struct rootsector	*rs;
	struct partition	*pi;
	
	if (dev_init( disk_device ) < 0)
		fatal( unable_to_open );
	if (dev_read( buffer, 0 ))
		fatal( unable_to_read );

	using_ICD_format = -1; /* don't know yet... */
	partitions = 0;
	rootsec_bootable = check_rootsec_checksum( buffer );
	
	rs = (struct rootsector *)buffer;
	hd_size = rs->hd_size;

	for( pi = &rs->part[0]; pi < &rs->part[4]; pi++ ) {

		if (pi->flag & PART_FLAG_VALID) {
			/* valid partition entry */
			if (PID_EQ( pi->id, "XGM" )) {
				/* extended part. -> cannot be ICD format */
				using_ICD_format = 0;
				if (i != &rs->part[3]) {
					fprintf( stderr, "This version of fdisk cannot handle "
							 "extended partitions rooted not to\n"
							 "the last primary partition :-(\n" );
					fatal( general );
				}
				read_extended( pi );
			}
			else {
				/* "normal" partiton */
				check_n_partitions();
				diskpart2PART( pi, 0, &part_table[partition] );
				++partition;
			}
		}
    }

	if (using_ICD_format == -1) {
		/* no extended partitions found -> test for ICD format */
		pi = &rs->icdpart[0];
		/* sanity check: no ICD format if first partion invalid */
		if ((pi->flag & PART_FLAG_VALID) && is_valid_ID( pi->id )) {
			using_ICD_format = 1;
			for ( ; pi < &rs->icdpart[8]; pi++ ) {
				if ((pi->flag & PART_FLAG_VALID) && is_valid_ID( pi->id )) {
					check_n_partitions();
					diskpart2PART( pi, 0, &part_table[partition] );
					++partition;
				}
			}
		}
	}

	if (using_ICD_format == -1)
		using_ICD_format = 0;
}


#define	MAGIC_CHECKSUM	0x1234

int check_rootsec_checksum( char *buf )

{	signed short *p, *pend = buf + SECTOR_SIZE, sum;

	for( sum = 0, p = (signed short *)buf; p < pend; ++p )
		sum += *p;
	return( sum == MAGIC_CHECKSUM );
}


void recalc_rootsec_checksum( char *buf )

{	signed short *p, *pend = buf + SECTOR_SIZE - 2, sum;

	for( sum = 0, p = (signed short *)buf; p < pend; ++p )
		sum += *p;
	*pend = MAGIC_CHECKSUM - sum;
}


void main( int argc, char *argv[] )

{
	if (argc > 3)
		fatal( usage );

	if (argc > 1 && *argv[1] == '-') {
		switch (*(argv[1] + 1)) {

		  case 'v':
			printf("This is Atari fdisk version " VERSION "\n");
			exit(0);

		  case 'l':
			listing = 1;
			try("/dev/hda");
			try("/dev/hdb");
			try("/dev/sda");
			try("/dev/sdb");
			try("/dev/sdc");
			try("/dev/sdd");
			try("/dev/sde");
			try("/dev/sdf");
			try("/dev/sdg");
			try("/dev/sdh");
			try("/dev/ada");
			try("/dev/adb");
			try("/dev/adc");
			try("/dev/add");
			try("/dev/ade");
			try("/dev/adf");
			try("/dev/adg");
			try("/dev/adh");
			exit( 0 );

		  case 's':
		  { int i;

			if (argc < 3)
				fatal( usage );
			if (strncmp( argv[2], "/dev/", 5 ) || !(i = atoi(argv[2] + 8)))
				fatal( usage );
			disk_device = (char *)malloc( 9 );
			strncpy( disk_device, argv[2], 8 );
			disk_device[8] = 0;
			if (dev_init( disk_device ) >= 0) {
				dev_deinit();
				get_boot();
				if (i > partitions) exit( 1 );
				printf("%d\n", part_table[i]->nr_sects / 2);
				exit( 0 );
			}
		  }
		  default:
			fatal( usage );
		}
	}
	
	if (argc > 1)
		disk_device = argv[argc - 1];
	else if (dev_init( DEFAULT_DEVICE ) < 0) {
		if (dev_init( disk_device = ALTERNATE_DEVICE ) < 0)
			disk_device = ALTERNATE2_DEVICE;
		else
			dev_deinit();
	}
	else
		dev_deinit();

	get_boot();
	if (argc == 1)
		printf( "Using %s as default device!\n", disk_device );

	while (1) {
		putchar( '\n' );
		switch( tolower(read_char("Command (m for help): "))) {
		  case 'a':
			set_bootable( get_partition( 1, partitions ));
			break;
		  case 'd':
			delete_partition( get_partition( 1, partitions ));
			break;
		  case 'l':
			list_types();
			break;
		  case 'n':
			new_partition();
			break;
		  case 'p':
			list_table();
			break;
		  case 'q':
			dev_deinit();
			exit( 0 );
		  case 't':
			change_partid();
			break;
		  case 'v':
			verify();
			break;
		  case 'w':
			write_table();
			break;
		  default:
			menu();
		}
	}
}


